use std::convert::From;
use std::env::args;
use std::fs::read_to_string;
use std::io::{Read, stdin};
use std::str::FromStr;
use std::process::exit;
use sha2::Sha256;
use hmac::{Hmac, Mac};
type HmacSha256 = Hmac<Sha256>;

mod helpers;

#[derive(Debug)]
pub enum Error {
    WrongKind(String),
    WrongClaim(String),
    IOError(std::io::Error)
}

impl From<std::io::Error> for Error {
    fn from(error: std::io::Error) -> Error {
        Error::IOError(error)
    }
}

pub enum SigType {
    SHA256HMAC,
}

pub enum Kind {
    Sig(String, SigType),
    Token,
}

impl FromStr for Kind {
    type Err = Error;
    fn from_str(s: &str) -> Result<Kind, Self::Err> {
        match s {
            "hmac-sha256" => {
                let mut body = String::new();
                let mut input = stdin();
                input.read_to_string(&mut body)?;
                println!("Body size: {}", body.len());
                Ok(Kind::Sig(body, SigType::SHA256HMAC))
            },
            "token" => {
                Ok(Kind::Token)
            },
            "help" | "-h" | "--help" => {
                help();
                exit(0);
            },
            _ => {
                Err(Error::WrongKind(s.to_string()))
            }
        }
    }
}

impl Kind {
    // Returns true if the check succeeded, false otherwise
    pub fn check(&self, secret: &str, claim: &str) -> bool {
        match self {
            Kind::Sig(body, sigtype) => {
                println!("|{}|", &body);
                match sigtype {
                    SigType::SHA256HMAC => {
                        let mut mac = HmacSha256::new_from_slice(secret.as_bytes()).unwrap();
                        mac.update(body.as_bytes());
                        let computed_sig = hex::encode(mac.finalize().into_bytes());
                        println!("Computed: {}", &computed_sig);
                        return computed_sig == claim;
                    },
                }
            },
            Kind::Token => {
                return secret == claim;
            },
        }
    }
}

fn help() {
    println!("whck KIND IDENTIFIER CLAIM (+ BODY as STDIN)");
    println!("  KIND: sig/token");
    println!("  IDENTIFIER: the unique identifier for the stored secret in $XDG_CONFIG_DIR/whck/ folder");
    println!("  CLAIM: the claimed sig/token to verify");
    println!("  BODY: the webhook payload, passed via STDIN");
    println!("");
    println!("If you're checking a signature, don't forget to pass the body as STDIN");
    println!("NOTE: If you're protecting secrets from other users (chmod -R 700 $XDG_CONFIG_HOME/whck),");
    println!("don't forget to set the suid bit on the program (chmod u+s $(which whck)).")
}

// Panic on malformed $XDG_CONFIG_DIR
fn read_secret(id: &str) -> String {
    if id.starts_with("/") || id.contains("/../") {
        println!("Malformed identifier");
        exit(1);
    }
    // $WHCK_DIR is used for tests
    // Otherwise, try $XDG_CONFIG_HOME/whck or $HOME/.config/whck
    // When home isn't defined try to check if $0 starts with /home and if so assume $HOME is there
    let mut secret_path = helpers::whck_dir();
    secret_path.push(id);
    let secret = read_to_string(&secret_path).expect(&format!("Failed to read secret from: {}", &secret_path.to_str().unwrap()));
    return secret;
}

fn main() -> Result<(), Error> {
    let mut cli_args = args();
    let arg_kind = cli_args.nth(1).expect("Missing argument: KIND");
    let kind = Kind::from_str(&arg_kind)?;
    let id = cli_args.nth(0).expect("Missing argument: IDENTIFIER");
    let claim = cli_args.nth(0).expect("Missing argument: CLAIM");

    let secret = read_secret(&id);
    if ! kind.check(&secret, &claim) {
        return Err(Error::WrongClaim(claim.clone()))
    }

    Ok(())
}
